"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.loadRemoteNxPlugin = void 0;
const child_process_1 = require("child_process");
const path = require("path");
// TODO (@AgentEnder): After scoped verbose logging is implemented, re-add verbose logs here.
// import { logger } from '../../utils/logger';
const internal_api_1 = require("./internal-api");
const messaging_1 = require("./messaging");
const cleanupFunctions = new Set();
const pluginNames = new Map();
function loadRemoteNxPlugin(plugin, root) {
    // this should only really be true when running unit tests within
    // the Nx repo. We still need to start the worker in this case,
    // but its typescript.
    const isWorkerTypescript = path.extname(__filename) === '.ts';
    const workerPath = path.join(__dirname, 'plugin-worker');
    const worker = (0, child_process_1.fork)(workerPath, [], {
        stdio: ['ignore', 'inherit', 'inherit', 'ipc'],
        env: {
            ...process.env,
            ...(isWorkerTypescript
                ? {
                    // Ensures that the worker uses the same tsconfig as the main process
                    TS_NODE_PROJECT: path.join(__dirname, '../../../tsconfig.lib.json'),
                }
                : {}),
        },
        execArgv: [
            ...process.execArgv,
            // If the worker is typescript, we need to register ts-node
            ...(isWorkerTypescript ? ['-r', 'ts-node/register'] : []),
        ],
    });
    worker.send((0, messaging_1.createMessage)({ type: 'load', payload: { plugin, root } }));
    // logger.verbose(`[plugin-worker] started worker: ${worker.pid}`);
    const pendingPromises = new Map();
    const exitHandler = createWorkerExitHandler(worker, pendingPromises);
    const cleanupFunction = () => {
        worker.off('exit', exitHandler);
        shutdownPluginWorker(worker, pendingPromises);
    };
    cleanupFunctions.add(cleanupFunction);
    return [
        new Promise((res, rej) => {
            worker.on('message', createWorkerHandler(worker, pendingPromises, res, rej));
            worker.on('exit', exitHandler);
        }),
        () => {
            cleanupFunction();
            cleanupFunctions.delete(cleanupFunction);
        },
    ];
}
exports.loadRemoteNxPlugin = loadRemoteNxPlugin;
async function shutdownPluginWorker(worker, pendingPromises) {
    // Clears the plugin cache so no refs to the workers are held
    internal_api_1.nxPluginCache.clear();
    // logger.verbose(`[plugin-pool] starting worker shutdown`);
    // Other things may be interacting with the worker.
    // Wait for all pending promises to be done before killing the worker
    await Promise.all(Array.from(pendingPromises.values()).map(({ promise }) => promise));
    worker.kill('SIGINT');
}
/**
 * Creates a message handler for the given worker.
 * @param worker Instance of plugin-worker
 * @param pending Set of pending promises
 * @param onload Resolver for RemotePlugin promise
 * @param onloadError Rejecter for RemotePlugin promise
 * @returns Function to handle messages from the worker
 */
function createWorkerHandler(worker, pending, onload, onloadError) {
    let pluginName;
    return function (message) {
        const parsed = JSON.parse(message);
        // logger.verbose(
        //   `[plugin-pool] received message: ${parsed.type} from ${
        //     pluginName ?? worker.pid
        //   }`
        // );
        (0, messaging_1.consumeMessage)(parsed, {
            'load-result': (result) => {
                if (result.success) {
                    const { name, createNodesPattern } = result;
                    pluginName = name;
                    pluginNames.set(worker, pluginName);
                    onload({
                        name,
                        createNodes: createNodesPattern
                            ? [
                                createNodesPattern,
                                (configFiles, ctx) => {
                                    const tx = pluginName + ':createNodes:' + performance.now();
                                    return registerPendingPromise(tx, pending, () => {
                                        worker.send((0, messaging_1.createMessage)({
                                            type: 'createNodes',
                                            payload: { configFiles, context: ctx, tx },
                                        }));
                                    });
                                },
                            ]
                            : undefined,
                        createDependencies: result.hasCreateDependencies
                            ? (opts, ctx) => {
                                const tx = pluginName + ':createDependencies:' + performance.now();
                                return registerPendingPromise(tx, pending, () => {
                                    worker.send((0, messaging_1.createMessage)({
                                        type: 'createDependencies',
                                        payload: { context: ctx, tx },
                                    }));
                                });
                            }
                            : undefined,
                        processProjectGraph: result.hasProcessProjectGraph
                            ? (graph, ctx) => {
                                const tx = pluginName + ':processProjectGraph:' + performance.now();
                                return registerPendingPromise(tx, pending, () => {
                                    worker.send((0, messaging_1.createMessage)({
                                        type: 'processProjectGraph',
                                        payload: { graph, ctx, tx },
                                    }));
                                });
                            }
                            : undefined,
                    });
                }
                else if (result.success === false) {
                    onloadError(result.error);
                }
            },
            createDependenciesResult: ({ tx, ...result }) => {
                const { resolver, rejector } = pending.get(tx);
                if (result.success) {
                    resolver(result.dependencies);
                }
                else if (result.success === false) {
                    rejector(result.error);
                }
            },
            createNodesResult: ({ tx, ...result }) => {
                const { resolver, rejector } = pending.get(tx);
                if (result.success) {
                    resolver(result.result);
                }
                else if (result.success === false) {
                    rejector(result.error);
                }
            },
            processProjectGraphResult: ({ tx, ...result }) => {
                const { resolver, rejector } = pending.get(tx);
                if (result.success) {
                    resolver(result.graph);
                }
                else if (result.success === false) {
                    rejector(result.error);
                }
            },
        });
    };
}
function createWorkerExitHandler(worker, pendingPromises) {
    return () => {
        for (const [_, pendingPromise] of pendingPromises) {
            pendingPromise.rejector(new Error(`Plugin worker ${pluginNames.get(worker) ?? worker.pid} exited unexpectedly with code ${worker.exitCode}`));
        }
    };
}
process.on('exit', () => {
    for (const fn of cleanupFunctions) {
        fn();
    }
});
function registerPendingPromise(tx, pending, callback) {
    let resolver, rejector;
    const promise = new Promise((res, rej) => {
        resolver = res;
        rejector = rej;
        callback();
    }).finally(() => {
        pending.delete(tx);
    });
    pending.set(tx, {
        promise,
        resolver,
        rejector,
    });
    return promise;
}
