"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.configurationGlobs = exports.createProjectConfigurations = exports.retrieveProjectConfigurationsWithoutPluginInference = exports.retrieveProjectConfigurationPaths = exports.retrieveProjectConfigurationsWithAngularProjects = exports.retrieveProjectConfigurations = exports.retrieveWorkspaceFiles = void 0;
const perf_hooks_1 = require("perf_hooks");
const angular_json_1 = require("../../adapter/angular-json");
const nx_json_1 = require("../../config/nx-json");
const project_configuration_utils_1 = require("./project-configuration-utils");
const internal_api_1 = require("../plugins/internal-api");
const workspace_context_1 = require("../../utils/workspace-context");
const build_all_workspace_files_1 = require("./build-all-workspace-files");
const path_1 = require("path");
/**
 * Walks the workspace directory to create the `projectFileMap`, `ProjectConfigurations` and `allWorkspaceFiles`
 * @throws
 * @param workspaceRoot
 * @param nxJson
 */
async function retrieveWorkspaceFiles(workspaceRoot, projectRootMap) {
    perf_hooks_1.performance.mark('native-file-deps:start');
    perf_hooks_1.performance.mark('native-file-deps:end');
    perf_hooks_1.performance.measure('native-file-deps', 'native-file-deps:start', 'native-file-deps:end');
    perf_hooks_1.performance.mark('get-workspace-files:start');
    const { projectFileMap, globalFiles, externalReferences } = (0, workspace_context_1.getNxWorkspaceFilesFromContext)(workspaceRoot, projectRootMap);
    perf_hooks_1.performance.mark('get-workspace-files:end');
    perf_hooks_1.performance.measure('get-workspace-files', 'get-workspace-files:start', 'get-workspace-files:end');
    return {
        allWorkspaceFiles: (0, build_all_workspace_files_1.buildAllWorkspaceFiles)(projectFileMap, globalFiles),
        fileMap: {
            projectFileMap,
            nonProjectFiles: globalFiles,
        },
        rustReferences: externalReferences,
    };
}
exports.retrieveWorkspaceFiles = retrieveWorkspaceFiles;
/**
 * Walk through the workspace and return `ProjectConfigurations`. Only use this if the projectFileMap is not needed.
 */
async function retrieveProjectConfigurations(plugins, workspaceRoot, nxJson) {
    const projects = await _retrieveProjectConfigurations(workspaceRoot, nxJson, plugins);
    return projects;
}
exports.retrieveProjectConfigurations = retrieveProjectConfigurations;
async function retrieveProjectConfigurationsWithAngularProjects(workspaceRoot, nxJson) {
    const pluginsToLoad = nxJson?.plugins ?? [];
    if ((0, angular_json_1.shouldMergeAngularProjects)(workspaceRoot, true) &&
        !pluginsToLoad.some((p) => p === angular_json_1.NX_ANGULAR_JSON_PLUGIN_NAME ||
            (typeof p === 'object' && p.plugin === angular_json_1.NX_ANGULAR_JSON_PLUGIN_NAME))) {
        pluginsToLoad.push((0, path_1.join)(__dirname, '../../adapter/angular-json'));
    }
    const [plugins, cleanup] = await (0, internal_api_1.loadNxPluginsInIsolation)(nxJson?.plugins ?? [], workspaceRoot);
    const res = _retrieveProjectConfigurations(workspaceRoot, nxJson, await plugins);
    cleanup();
    return res;
}
exports.retrieveProjectConfigurationsWithAngularProjects = retrieveProjectConfigurationsWithAngularProjects;
function _retrieveProjectConfigurations(workspaceRoot, nxJson, plugins) {
    const globPatterns = configurationGlobs(plugins);
    const projectFiles = (0, workspace_context_1.globWithWorkspaceContext)(workspaceRoot, globPatterns);
    return createProjectConfigurations(workspaceRoot, nxJson, projectFiles, plugins);
}
function retrieveProjectConfigurationPaths(root, plugins) {
    const projectGlobPatterns = configurationGlobs(plugins);
    return (0, workspace_context_1.globWithWorkspaceContext)(root, projectGlobPatterns);
}
exports.retrieveProjectConfigurationPaths = retrieveProjectConfigurationPaths;
const projectsWithoutPluginCache = new Map();
// TODO: This function is called way too often, it should be optimized without this cache
async function retrieveProjectConfigurationsWithoutPluginInference(root) {
    const nxJson = (0, nx_json_1.readNxJson)(root);
    const [plugins, cleanup] = await (0, internal_api_1.loadNxPluginsInIsolation)([]); // only load default plugins
    const projectGlobPatterns = retrieveProjectConfigurationPaths(root, plugins);
    const cacheKey = root + ',' + projectGlobPatterns.join(',');
    if (projectsWithoutPluginCache.has(cacheKey)) {
        return projectsWithoutPluginCache.get(cacheKey);
    }
    const projectFiles = (0, workspace_context_1.globWithWorkspaceContext)(root, projectGlobPatterns);
    const { projects } = await createProjectConfigurations(root, nxJson, projectFiles, plugins);
    projectsWithoutPluginCache.set(cacheKey, projects);
    cleanup();
    return projects;
}
exports.retrieveProjectConfigurationsWithoutPluginInference = retrieveProjectConfigurationsWithoutPluginInference;
async function createProjectConfigurations(workspaceRoot, nxJson, configFiles, plugins) {
    perf_hooks_1.performance.mark('build-project-configs:start');
    const { projects, externalNodes, rootMap, sourceMaps } = await (0, project_configuration_utils_1.buildProjectsConfigurationsFromProjectPathsAndPlugins)(nxJson, configFiles, plugins, workspaceRoot);
    perf_hooks_1.performance.mark('build-project-configs:end');
    perf_hooks_1.performance.measure('build-project-configs', 'build-project-configs:start', 'build-project-configs:end');
    return {
        projects,
        externalNodes,
        projectRootMap: rootMap,
        sourceMaps,
    };
}
exports.createProjectConfigurations = createProjectConfigurations;
function configurationGlobs(plugins) {
    const globPatterns = [];
    for (const plugin of plugins) {
        if ('createNodes' in plugin && plugin.createNodes) {
            globPatterns.push(plugin.createNodes[0]);
        }
    }
    return globPatterns;
}
exports.configurationGlobs = configurationGlobs;
